@@ -27,7 +27,7 @@
 #define NCURSES_VERSION "1.9.4"
 
 #include <stdio.h>   
-#include <unctrl.h>
+#include <ncurses/unctrl.h>
 #include <stdarg.h>
 #ifdef XPG4_EXTENDED
 #include <stddef.h>	/* we want wchar_t */
