@@ -19,15 +19,31 @@ library source.
 #include <stdio.h>
 #include <ctype.h>
 #include <stdlib.h>
+#include <unistd.h>
 #include <string.h>
 #include <assert.h>
-#include <ncurses.h>
+#include <curses.h>
+#include <panel.h>
+#include <menu.h>
 
-#define P(s)		printw("%s\n", s)
+#define P(string)	printw("%s\n", string)
 #ifndef CTRL
 #define CTRL(x)		((x) & 0x1f)
 #endif
 
+static void Pause(void)
+{
+	move(LINES - 1, 0);
+	addstr("Press any key to continue... ");
+	(void) getch();
+}
+
+static void Cannot(char *what)
+{
+	printw("\nThis %s terminal %s\n\n", getenv("TERM"), what);
+	Pause();
+}
+
 /****************************************************************************
  *
  * Character input test
@@ -37,13 +53,28 @@ library source.
 static void getch_test(void)
 /* test the keypad feature */
 {
-    unsigned int c;
-    int firsttime = 0;
-
-    refresh();
-
-    c = '?';
-    do {
+char buf[BUFSIZ];
+int c;
+int incount = 0, firsttime = 0;
+bool blocking = TRUE;
+int y, x;
+  
+     refresh();
+  
+     (void) printw("Delay in 10ths of a second (<CR> for blocking input)? ");
+     echo();
+     getstr(buf);
+     noecho();
+ 
+     if (isdigit(buf[0]))
+     {
+ 	timeout(atoi(buf) * 100);
+ 	blocking = FALSE;
+     }
+ 
+     c = '?';
+     for (;;)
+     {
 	if (firsttime++)
 	{
 	    printw("Key pressed: %04o ", c);
@@ -54,10 +85,11 @@ static void getch_test(void)
 	    }
 	    else if (c > 0x80)
 	    {
-		if (isprint(c & ~0x80))
-		    (void) printw("M-%c", c);
+		int c2 = (c & 0x7f);
+		if (isprint(c2))
+		    (void) printw("M-%c", c2);
 		else
-		    (void) printw("M-%s", unctrl(c));
+		    (void) printw("M-%s", unctrl(c2));
 		addstr(" (high-half character)\n");
 	    }
 	    else
@@ -67,20 +99,50 @@ static void getch_test(void)
 		else
 		    (void) printw("%s (ASCII control character)\n", unctrl(c));
 	    }
+	    getyx(stdscr, y, x);
+	    if (y >= LINES-1)
+	    	move(0,0);
+	    clrtoeol();	
+	}
+
+	if (c == 'g')
+	{
+	    addstr("getstr test: ");
+	    echo(); getstr(buf); noecho();
+	    printw("I saw `%s'.\n", buf);
+	}
+	if (c == 's')
+	{
+	    addstr("Shelling out...");
+	    def_prog_mode();
+	    endwin();
+	    system("sh");
+	    addstr("returned from shellout.\n");
+	    refresh();
 	}
 	if (c == 'x' || c == 'q')
 	    break;
 	if (c == '?')
-	    addstr("Type any key to see its keypad value, `q' to quit, `?' for help.\n");
-    } while
-	((c = getch()) != EOF);
+	{
+	    addstr("Type any key to see its keypad value.  Also:\n");
+	    addstr("g -- triggers a getstr test\n");
+	    addstr("s -- shell out\n");
+	    addstr("q -- quit\n");
+	    addstr("? -- repeats this help message\n");
+	}
 
+	while ((c = getch()) == ERR)
+	    if (!blocking)
+		(void) printw("%05d: input timed out\n", incount++);
+    }
+
+    timeout(-1);
     erase();
     endwin();
 }
 
 static void attr_test(void)
-/* test text attrivutes */
+/* test text attributes */
 {
     refresh();
 
@@ -88,48 +150,50 @@ static void attr_test(void)
 
     mvaddstr(2,8,"This is STANDOUT mode: ");
     attron(A_STANDOUT);
-    addstr("abcde fghij klmno pqrst uvwxy x");
+    addstr("abcde fghij klmno pqrst uvwxy z");
     attroff(A_STANDOUT);
 
     mvaddstr(4,8,"This is REVERSE mode: ");
     attron(A_REVERSE);
-    addstr("abcde fghij klmno pqrst uvwxy x");
+    addstr("abcde fghij klmno pqrst uvwxy z");
     attroff(A_REVERSE);
 
     mvaddstr(6,8,"This is BOLD mode: ");
     attron(A_BOLD);
-    addstr("abcde fghij klmno pqrst uvwxy x");
+    addstr("abcde fghij klmno pqrst uvwxy z");
     attroff(A_BOLD);
 
     mvaddstr(8,8,"This is UNDERLINE mode: ");
     attron(A_UNDERLINE);
-    addstr("abcde fghij klmno pqrst uvwxy x");
+    addstr("abcde fghij klmno pqrst uvwxy z");
     attroff(A_UNDERLINE);
 
     mvaddstr(10,8,"This is DIM mode: ");
     attron(A_DIM);
-    addstr("abcde fghij klmno pqrst uvwxy x");
+    addstr("abcde fghij klmno pqrst uvwxy z");
     attroff(A_DIM);
 
     mvaddstr(12,8,"This is BLINK mode: ");
     attron(A_BLINK);
-    addstr("abcde fghij klmno pqrst uvwxy x");
+    addstr("abcde fghij klmno pqrst uvwxy z");
     attroff(A_BLINK);
 
-    mvaddstr(14,8,"This is BOLD UNDERLINE BLINK mode: ");
-    attron(A_BOLD|A_BLINK|A_UNDERLINE);
-    addstr("abcde fghij klmno pqrst uvwxy x");
-    attroff(A_BOLD|A_BLINK|A_UNDERLINE);
+    mvaddstr(14,8,"This is PROTECT mode: ");
+    attron(A_PROTECT);
+    addstr("abcde fghij klmno pqrst uvwxy z");
+    attroff(A_PROTECT);
+
+    attrset(A_INVIS);
+    mvaddstr(16,8,"This is INVISIBLE mode: ");
+    addstr("abcde fghij klmno pqrst uvwxy z");
 
     attrset(A_NORMAL);
-    mvaddstr(16,8,"This is NORMAL mode: ");
-    addstr("abcde fghij klmno pqrst uvwxy x");
+    mvaddstr(18,8,"This is NORMAL mode: ");
+    addstr("abcde fghij klmno pqrst uvwxy z");
 
     refresh();
 
-    move(LINES - 1, 0);
-    addstr("Press any key to continue... ");
-    (void) getch();
+    Pause();
 
     erase();
     endwin();
@@ -171,7 +235,7 @@ static void color_test(void)
     for (i = 1; i < COLOR_PAIRS; i++)
     {
 	init_pair(i, i % COLORS, i / COLORS);
-	attron(COLOR_PAIR(i));
+	attron((attr_t)COLOR_PAIR(i));
 	mvaddstr(3 + (i / COLORS), (i % COLORS + 1) * 8, "Hello");
 	attrset(A_NORMAL);
     }
@@ -186,14 +250,12 @@ static void color_test(void)
     for (i = 1; i < COLOR_PAIRS; i++)
     {
 	init_pair(i, i % COLORS, i / COLORS);
-	attron(COLOR_PAIR(i) | A_BOLD);
+	attron((attr_t)(COLOR_PAIR(i) | A_BOLD));
 	mvaddstr(6 + COLORS + (i / COLORS), (i % COLORS + 1) * 8, "Hello");
 	attrset(A_NORMAL);
     }
 
-    move(LINES - 1, 0);
-    addstr("Press any key to continue... ");
-    (void) getch();
+    Pause();
 
     erase();
     endwin();
@@ -329,9 +391,7 @@ static void color_edit(void)
     P("");
     P("To quit, do `x' or 'q'");
 
-	    move(LINES - 1, 0);
-	    addstr("Press any key to continue... ");
-	    (void) getch();
+	    Pause();
 	    erase();
 	    break;
 
@@ -403,8 +463,8 @@ static void slk_test(void)
 
 	case 's':
 	    move(20, 0);
-	    while ((c = getch()) != 'Q')
-		addch(c);
+	    while ((c = getch()) != 'Q' && (c != ERR))
+		addch((chtype)c);
 	    break;
 
 	case 'd':
@@ -426,7 +486,9 @@ static void slk_test(void)
 	case '1': case '2': case '3': case '4':
 	case '5': case '6': case '7': case '8':
 	    (void) mvaddstr(20, 0, "Please enter the label value: ");
+	    echo();
 	    wgetnstr(stdscr, buf, 8);
+	    noecho();
 	    slk_set((c - '0'), buf, fmt);
 	    slk_refresh();
 	    break;
@@ -452,17 +514,18 @@ static void slk_test(void)
  *
  ****************************************************************************/
 
-static void acs_display()
+static void acs_display(void)
 /* display the ACS character set */
 {
-traceon();
+    int	i, j;
+
     erase();
     attron(A_BOLD);
     mvaddstr(0, 20, "Display of the ACS Character Set");
     attroff(A_BOLD);
     refresh();
 
-#define ACSY	5
+#define ACSY	2
     mvaddstr(ACSY + 0, 0, "ACS_ULCORNER: "); addch(ACS_ULCORNER);
     mvaddstr(ACSY + 1, 0, "ACS_LLCORNER: "); addch(ACS_LLCORNER);
     mvaddstr(ACSY + 2, 0, "ACS_URCORNER: "); addch(ACS_URCORNER);
@@ -490,9 +553,16 @@ traceon();
     mvaddstr(ACSY + 10,40, "ACS_LANTERN: "); addch(ACS_LANTERN);
     mvaddstr(ACSY + 11,40, "ACS_BLOCK: "); addch(ACS_BLOCK);
 
-    move(LINES - 1, 0);
-    addstr("Press any key to continue... ");
-    (void) getch();
+#define HYBASE 	(ACSY + 13)    
+    mvprintw(HYBASE + 1, 0, "High-half characters via echochar:\n");
+    for (i = 0; i < 4; i++)
+    {
+	move(HYBASE + i + 3, 24);
+	for (j = 0; j < 32; j++)
+	    echochar((chtype)(128 + 32 * i + j));
+    }
+
+    Pause();
 
     erase();
     endwin();
@@ -504,6 +574,8 @@ traceon();
  *
  ****************************************************************************/
 
+#define BOTLINES	3	/* number of line stolen from screen bottom */
+
 typedef struct
 {
     int y, x;
@@ -521,9 +593,8 @@ static void report(void)
     move(y, x);
 }
 
-static pair *selectcell(uli, ulj, lri, lrj)
+static pair *selectcell(int uli, int ulj, int lri, int lrj)
 /* arrows keys move cursor, return location at current on non-arrow key */
-int	uli, ulj, lri, lrj;	/* co-ordinates of corners */
 {
     static pair	res;			/* result cell */
     int		si = lri - uli + 1;	/* depth of the select area */
@@ -542,6 +613,7 @@ int	uli, ulj, lri, lrj;	/* co-ordinates of corners */
 	case KEY_DOWN:	i++; break;
 	case KEY_LEFT:	j += sj - 1; break;
 	case KEY_RIGHT:	j++; break;
+	case '\004':	return((pair *)NULL);
 	default:	res.y = uli + i; res.x = ulj + j; return(&res);
 	}
 	i %= si;
@@ -552,26 +624,33 @@ int	uli, ulj, lri, lrj;	/* co-ordinates of corners */
 static WINDOW *getwindow(void)
 /* Ask user for a window definition */
 {
-    WINDOW	*rwindow, *bwindow;
-    pair	ul, lr;
+    WINDOW	*rwindow;
+    pair	ul, lr, *tmp;
 
     move(0, 0); clrtoeol();
     addstr("Use arrows to move cursor, anything else to mark corner 1");
     refresh();
-    memcpy(&ul, selectcell(1,    0,    LINES-1, COLS-1), sizeof(pair));
-    addch(ACS_ULCORNER);
+    if ((tmp = selectcell(1,    0,    LINES-BOTLINES, COLS-1)) == (pair *)NULL)
+	return((WINDOW *)NULL);
+    memcpy(&ul, tmp, sizeof(pair));
+    mvaddch(ul.y-1, ul.x-1, ACS_ULCORNER);
     move(0, 0); clrtoeol();
     addstr("Use arrows to move cursor, anything else to mark corner 2");
     refresh();
-    memcpy(&lr, selectcell(ul.y, ul.x, LINES-1, COLS-1), sizeof(pair));
+    if ((tmp = selectcell(ul.y, ul.x, LINES-BOTLINES, COLS-1)) == (pair *)NULL)
+	return((WINDOW *)NULL);
+    memcpy(&lr, tmp, sizeof(pair));
 
-    rwindow = newwin(lr.y - ul.y + 1, lr.x - ul.x + 1, ul.y, ul.x);
+    rwindow = subwin(stdscr, lr.y - ul.y + 1, lr.x - ul.x + 1, ul.y, ul.x);
 
-    bwindow = newwin(lr.y - ul.y + 3, lr.x - ul.x + 3, ul.y - 1, ul.x - 1);
-    wborder(bwindow, ACS_VLINE, ACS_VLINE, ACS_HLINE, ACS_HLINE,
-		0, 0, 0, 0);
-    wrefresh(bwindow);
-    delwin(bwindow);
+    mvaddch(ul.y-1, lr.x+1, ACS_URCORNER);
+    mvaddch(lr.y+1, lr.x+1, ACS_LRCORNER);
+    mvaddch(lr.y+1, ul.x-1, ACS_LLCORNER);
+    move(ul.y-1, ul.x);   hline(ACS_HLINE, lr.x - ul.x + 1);
+    move(ul.y,   ul.x-1); vline(ACS_VLINE, lr.y - ul.y + 1);
+    move(lr.y+1, ul.x);   hline(ACS_HLINE, lr.x - ul.x + 1);
+    move(ul.y,   lr.x+1); vline(ACS_VLINE, lr.y - ul.y + 1);
+    refresh();
 
     scrollok(rwindow, TRUE);
 /*    immedok(rwindow);	*/
@@ -580,22 +659,42 @@ static WINDOW *getwindow(void)
     return(rwindow);
 }
 
-static void acs_and_scroll()
+static void transient(char *msg)
+{
+    if (msg)
+    {
+	mvaddstr(LINES - 1, 0, msg);
+	refresh();
+	sleep(1);
+    }
+
+    mvaddstr(LINES - 1, 0,
+	     "All other characters are echoed, windows should scroll.");
+    refresh();
+}
+
+static void acs_and_scroll(void)
 /* Demonstrate windows */
 {
     int	c;
+    FILE *fp;
     struct frame
     {
         struct frame	*next, *last;
         WINDOW		*wind;
     }
-    *oldw  = (struct frame *)NULL, *current = (struct frame *)NULL, *neww;
+    *current = (struct frame *)NULL, *neww;
+
+#define DUMPFILE	"screendump"
 
     refresh();
+    mvaddstr(LINES - 4, 0,
+	     "F1 = make new window, F2 = next window, F3 = previous window, ");
+    mvaddstr(LINES - 3, 0,
+	     "F4 = scroll current window forward, F5 = scroll current window backward");
     mvaddstr(LINES - 2, 0,
-	     "F1 = make new window, F2 = next window, F3 = previous window");
-    mvaddstr(LINES - 1, 0,
-	     "All other characters are echoed, windows should scroll.");
+	     "F6 = save window to file, F7 = restore window, Ctrl-D = exit");
+    transient((char *)NULL);
 
     c = KEY_F(1);
     do {
@@ -607,31 +706,84 @@ static void acs_and_scroll()
 	{
 	case KEY_F(1):
 	    neww = (struct frame *) malloc(sizeof(struct frame));
-	    neww->wind = getwindow();
-	    if (oldw == NULL)	/* First element,  */
+	    if ((neww->wind = getwindow()) == (WINDOW *)NULL)
+		goto breakout;
+
+	    if (current == NULL)	/* First element,  */
 	    {
 		neww->next = neww; /*   so point it at itself */
 		neww->last = neww;
-		current = neww;
 	    }
 	    else
 	    {
-		neww->last = oldw;  oldw->next = neww;
-		neww->next = current; current->last = neww;
+		neww->next = current->next;
+		neww->last = current;
+		neww->last->next = neww;
+		neww->next->last = neww;
 	    }
-	    oldw = neww;
+	    current = neww;
 	    keypad(neww->wind, TRUE);
 	    break;
 
-	case KEY_F(2):
+	case KEY_F(2):		/* go to next window */
 	    current = current->next;
 	    break;
 
-	case KEY_F(3):
+	case KEY_F(3):		/* go to previous window */
 	    current = current->last;
 	    break;
 
-	case KEY_F(4):	/* undocumented --- use this to test area clears */
+	case KEY_F(4):		/* scroll current window forward */
+	    if (current)
+		wscrl(current->wind, 1);
+	    break;
+
+	case KEY_F(5):		/* scroll current window backwards */
+	    if (current)
+		wscrl(current->wind, -1);
+	    break;
+
+	case KEY_F(6):		/* save window */
+	    if ((fp = fopen(DUMPFILE, "w")) == (FILE *)NULL)
+		transient("Can't open screen dump file");
+	    else
+	    {
+		(void) putwin(current->wind, fp);
+		(void) fclose(fp);
+
+		current->last->next = current->next;
+		current->next->last = current->last;
+
+		werase(current->wind);
+		wrefresh(current->wind);
+		delwin(current->wind);
+
+		neww = current->next;
+		free(current);
+		current = neww;
+	    }
+	    break;
+
+	case KEY_F(7):		/* restore window */
+	    if ((fp = fopen(DUMPFILE, "r")) == (FILE *)NULL)
+		transient("Can't open screen dump file");
+	    else
+	    {
+		neww = (struct frame *) malloc(sizeof(struct frame));
+
+		neww->next = current->next;
+		neww->last = current;
+		neww->last->next = neww;
+		neww->next->last = neww;
+
+		neww->wind = getwin(fp);
+		(void) fclose(fp);
+
+		wrefresh(neww->wind);
+	    }
+	    break;
+
+	case KEY_F(10):	/* undocumented --- use this to test area clears */
 	    selectcell(0, 0, LINES - 1, COLS - 1);
 	    clrtobot();
 	    refresh();
@@ -642,51 +794,544 @@ static void acs_and_scroll()
 	    /* FALLTHROUGH */
 
 	default:
-	    waddch(current->wind, c);
+	    waddch(current->wind, (chtype)c);
 	    break;
 	}
 	report();
 	wrefresh(current->wind);
     } while
-	((c = wgetch(current->wind)) != '\004');
+	((c = wgetch(current->wind)) != '\004'
+	 && (c != ERR));
+
+ breakout:
+    erase();
+    endwin();
+}
+
+/****************************************************************************
+ *
+ * Panels tester
+ *
+ ****************************************************************************/
+
+static PANEL *p1;
+static PANEL *p2;
+static PANEL *p3;
+static PANEL *p4;
+static PANEL *p5;
+static WINDOW *w1;
+static WINDOW *w2;
+static WINDOW *w3;
+static WINDOW *w4;
+static WINDOW *w5;
+
+#define	nap(x)		usleep(1000*x)
+
+static unsigned long nap_msec = 1;
+
+char *mod[] = 
+{
+	"test ",
+	"TEST ",
+	"(**) ",
+	"*()* ",
+	"<--> ",
+	"LAST "
+};
+
+/*+-------------------------------------------------------------------------
+	wait_a_while(msec)
+--------------------------------------------------------------------------*/
+static void
+wait_a_while(unsigned long msec)
+{
+#ifdef NONAP
+	if(nap_msec == 1)
+		getchar();
+	else if(msec > 1000L)
+		sleep((int)msec/1000L);
+	else
+		sleep(1);
+#else
+	if(nap_msec == 1)
+		getchar();
+	else
+		nap(nap_msec);
+#endif
+}	/* end of wait_a_while */
+
+/*+-------------------------------------------------------------------------
+	saywhat(text)
+--------------------------------------------------------------------------*/
+static void
+saywhat(char *text)
+{
+	wmove(stdscr,LINES - 1,0);
+	wclrtoeol(stdscr);
+	waddstr(stdscr, text);
+}	/* end of saywhat */
+
+/*+-------------------------------------------------------------------------
+	mkpanel(rows,cols,tly,tlx) - alloc a win and panel and associate them
+--------------------------------------------------------------------------*/
+static PANEL *
+mkpanel(int rows, int cols, int tly, int tlx)
+{
+WINDOW *win = newwin(rows,cols,tly,tlx);
+PANEL *pan;
+
+	if(!win)
+		return((PANEL *)0);
+	if((pan = new_panel(win)))
+		return(pan);
+	delwin(win);
+	return((PANEL *)0);
+}	/* end of mkpanel */
+
+/*+-------------------------------------------------------------------------
+	rmpanel(pan)
+--------------------------------------------------------------------------*/
+static void
+rmpanel(PANEL *pan)
+{
+WINDOW *win = pan->win;
+	del_panel(pan);
+	delwin(win);
+}	/* end of rmpanel */
+
+/*+-------------------------------------------------------------------------
+	pflush()
+--------------------------------------------------------------------------*/
+static void
+pflush(void)
+{
+	update_panels();
+	doupdate();
+}	/* end of pflush */
+
+/*+-------------------------------------------------------------------------
+	fill_panel(win)
+--------------------------------------------------------------------------*/
+static void
+fill_panel(PANEL *pan)
+{
+WINDOW *win = pan->win;
+chtype num = *(pan->user + 1);
+int y,x;
+
+	box(win, 0, 0);  
+	wmove(win,1,1);
+	wprintw(win,"-pan%c-",num);
+	for(y = 2; y < getmaxy(win) - 1; y++)
+	{
+		for(x = 1; x < getmaxx(win) - 1; x++)
+		{
+			wmove(win,y,x);
+			waddch(win,num);
+		}
+	}
+}	/* end of fill_panel */
+
+/*+-------------------------------------------------------------------------
+	main(argc,argv)
+--------------------------------------------------------------------------*/
+
+static void demo_panels(void)
+{
+int itmp;
+register y,x;
+
+#ifdef FOO
+	if((argc > 1) && atol(argv[1]))
+		nap_msec = atol(argv[1]);
+#endif /* FOO */
+
+	refresh();
+
+	for(y = 0; y < LINES - 1; y++)
+	{
+		for(x = 0; x < COLS; x++)
+			wprintw(stdscr,"%d",(y + x) % 10);
+	}
+	for(y = 0; y < 5; y++)
+	{
+		p1 = mkpanel(10,10,0,0);
+		w1 = panel_window(p1);
+		set_panel_userptr(p1,"p1");
+
+		p2 = mkpanel(14,14,5,5);
+		w2 = panel_window(p2);
+		set_panel_userptr(p2,"p2");
+
+		p3 = mkpanel(6,8,12,12);
+		w3 = panel_window(p3);
+		set_panel_userptr(p3,"p3");
+
+		p4 = mkpanel(10,10,10,30);
+		w4 = panel_window(p4);
+		set_panel_userptr(p4,"p4");
+
+		p5 = mkpanel(10,10,13,37);
+		w5 = panel_window(p5);
+		set_panel_userptr(p5,"p5");
+
+		fill_panel(p1);
+		fill_panel(p2);
+		fill_panel(p3);
+		fill_panel(p4);
+		fill_panel(p5);
+		hide_panel(p4);
+		hide_panel(p5);
+		pflush();
+		saywhat("press any key to continue");
+		wait_a_while(nap_msec);
+
+		saywhat("h3 s1 s2 s4 s5; press any key to continue");
+		move_panel(p1,0,0);
+		hide_panel(p3);
+		show_panel(p1);
+		show_panel(p2);
+		show_panel(p4);
+		show_panel(p5);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("s1; press any key to continue");
+		show_panel(p1);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("s2; press any key to continue");
+		show_panel(p2);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("m2; press any key to continue");
+		move_panel(p2,10,10);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("s3;");
+		show_panel(p3);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("m3; press any key to continue");
+		move_panel(p3,5,5);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("b3; press any key to continue");
+		bottom_panel(p3);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("s4; press any key to continue");
+		show_panel(p4);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("s5; press any key to continue");
+		show_panel(p5);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("t3; press any key to continue");
+		top_panel(p3);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("t1; press any key to continue");
+		top_panel(p1);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("t2; press any key to continue");
+		top_panel(p2);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("t3; press any key to continue");
+		top_panel(p3);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("t4; press any key to continue");
+		top_panel(p4);
+		pflush();
+		wait_a_while(nap_msec);
+
+		for(itmp = 0; itmp < 6; itmp++)
+		{
+			saywhat("m4; press any key to continue");
+			wmove(w4,3,1);
+			waddstr(w4,mod[itmp]);
+			move_panel(p4,4,itmp*10);
+			wmove(w5,4,1);
+			waddstr(w5,mod[itmp]);
+			pflush();
+			wait_a_while(nap_msec);
+			saywhat("m5; press any key to continue");
+			wmove(w4,4,1);
+			waddstr(w4,mod[itmp]);
+			move_panel(p5,7,(itmp*10) + 6);
+			wmove(w5,3,1);
+			waddstr(w5,mod[itmp]);
+			pflush();
+			wait_a_while(nap_msec);
+		}
+
+		saywhat("m4; press any key to continue");
+		move_panel(p4,4,itmp*10);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("t5; press any key to continue");
+		top_panel(p5);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("t2; press any key to continue");
+		top_panel(p2);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("t1; press any key to continue");
+		top_panel(p1);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("d2; press any key to continue");
+		rmpanel(p2);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("h3; press any key to continue");
+		hide_panel(p3);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("d1; press any key to continue");
+		rmpanel(p1);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("d4; press any key to continue");
+		rmpanel(p4);
+		pflush();
+		wait_a_while(nap_msec);
+
+		saywhat("d5; press any key to continue");
+		rmpanel(p5);
+		pflush();
+		wait_a_while(nap_msec);
+		if(nap_msec == 1)
+			break;
+		nap_msec = 100L;
+	}
 
     erase();
     endwin();
 }
 
+/****************************************************************************
+ *
+ * Pad tester
+ *
+ ****************************************************************************/
+
+#define GRIDSIZE	5
+
+static void panner(WINDOW *pad, int iy, int ix, int (*pgetc)(WINDOW *))
+{
+    static int porty, portx, basex = 0, basey = 0;
+    int pxmax, pymax, lowend, highend, i, j, c;
+
+    porty = iy; portx = ix;
+
+    getmaxyx(pad, pymax, pxmax);
+
+    c = KEY_REFRESH;
+    do {
+	switch(c)
+	{
+	case KEY_REFRESH:
+	    /* do nothing */
+	    break;
+
+	case KEY_IC:
+	    if (portx >= pxmax || portx >= ix)
+		beep();
+	    else
+		++portx;
+	    break;
+
+	case KEY_IL:
+	    if (porty >= pymax || porty >= iy)
+		beep();
+	    else
+		++porty;
+	    break;
+
+	case KEY_DC:
+	    if (portx <= 0)
+		beep();
+	    else
+	    {
+		--portx;
+		for (i = 0; i < porty; i++)
+		    mvaddch(i, portx, ' ');
+	    }
+	    break;
+
+	case KEY_DL:
+	    if (porty <= 0)
+		beep();
+	    else
+	    {
+		--porty;
+		for (j = 0; j < portx; j++)
+		    mvaddch(porty, j, ' ');
+	    }
+	    break;
+
+	case KEY_LEFT:
+	    if (basex > 0)
+		basex--;
+	    else
+		beep();
+	    break;
+
+	case KEY_RIGHT:
+	    if (basex + portx - (pymax > porty) < pxmax)
+		basex++;
+	    else
+		beep();
+	    break;
+
+	case KEY_UP:
+	    if (basey > 0)
+		basey--;
+	    else
+		beep();
+	    break;
+
+	case KEY_DOWN:
+	    if (basey + porty - (pxmax > portx) < pymax)
+		basey++;
+	    else
+		beep();
+	    break;
+	}
+
+	if (pxmax > portx - 1) {
+
+	    lowend = basex * ((float)(portx - 1) / (float)pxmax);
+	    highend = (basex + portx - 1) * ((float)(portx - 1) / (float)pxmax);
+
+	    for (j = 0; j < lowend; j++)
+		mvaddch(porty - 1, j, ACS_HLINE);
+	    attron(A_REVERSE);
+	    for (j = lowend; j <= highend; j++)
+		mvaddch(porty - 1, j, ' ');
+	    attroff(A_REVERSE);
+	    for (j = highend + 1; j < portx; j++)
+		mvaddch(porty - 1, j, ACS_HLINE);
+        }
+	if (pymax > porty - 1) {
+
+	    lowend = basey * ((float)(porty - 1) / (float)pymax);
+	    highend = (basey + porty - 1) * ((float)(porty - 1) / (float)pymax);
+
+	    for (i = 0; i < lowend; i++)
+		mvaddch(i, portx - 1, ACS_VLINE);
+	    attron(A_REVERSE);
+	    for (i = lowend; i <= highend; i++)
+		mvaddch(i, portx - 1, ' ');
+	    attroff(A_REVERSE);
+	    for (i = highend + 1; i < porty; i++)
+		mvaddch(i, portx - 1, ACS_VLINE);
+        }
+
+	mvaddch(porty - 1, portx - 1, ACS_LRCORNER);
+	wnoutrefresh(stdscr);
+
+	prefresh(pad,
+		 basey, basex,
+		 0, 0,
+		 porty - (pxmax > portx) - 1,
+		 portx - (pymax > porty) - 1); 
+
+	doupdate();
+
+    } while
+	((c = pgetc(pad)) != KEY_EXIT);
+}
+
+static
+int padgetch(WINDOW *win)
+{
+    int	c;
+
+    switch(c = wgetch(win))
+    {
+    case 'u': return(KEY_UP);
+    case 'd': return(KEY_DOWN);
+    case 'r': return(KEY_RIGHT);
+    case 'l': return(KEY_LEFT);
+    case '+': return(KEY_IL);
+    case '-': return(KEY_DL);
+    case '>': return(KEY_IC);
+    case '<': return(KEY_DC);
+    case 'q': return(KEY_EXIT);
+    default: return(c);
+    }
+}
+
+static void demo_pad(void)
+/* Demonstrate pads. */
+{
+    int i, j;
+    unsigned gridcount = 0;
+    WINDOW *panpad = newpad(200, 200);
+
+    for (i = 0; i < 200; i++)
+    {
+	for (j = 0; j < 200; j++)
+	    if (i % GRIDSIZE == 0 && j % GRIDSIZE == 0)
+	    {
+		if (i == 0 || j == 0)
+		    waddch(panpad, '+');
+		else
+		    waddch(panpad, (chtype)('A' + (gridcount++ % 26)));
+	    }
+    	    else if (i % GRIDSIZE == 0)
+		waddch(panpad, '-');
+    	    else if (j % GRIDSIZE == 0)
+		waddch(panpad, '|');
+	    else
+		waddch(panpad, ' ');
+    }
+    mvprintw(LINES - 3, 0, "Use arrow keys to pan over the test pattern");
+    mvprintw(LINES - 2, 0, "Use +,- to grow/shrink the panner vertically.");
+    mvprintw(LINES - 1, 0, "Use <,> to grow/shrink the panner horizontally.");
+
+    keypad(panpad, TRUE);
+    panner(panpad, LINES - 4, COLS, padgetch);
+
+    endwin();
+    erase();
+}
+
 /****************************************************************************
  *
  * Tests from John Burnell's PDCurses tester
  *
  ****************************************************************************/
 
-static void demo_pad(void)
-/* Demonstrate pads. */
-{
-    WINDOW *pad;
-
-    pad = newpad(50,100);
-    mvwaddstr(pad, 5, 2, "This is a new pad");
-    mvwaddstr(pad, 8, 0, "The end of this line should be truncated here:abcd");
-    mvwaddstr(pad,11, 1, "This line should not appear.");
-    wmove(pad, 10, 1);
-    wclrtoeol(pad);
-    mvwaddstr(pad, 10, 1, " Press any key to continue");
-    prefresh(pad,0,0,0,0,10,45);
-    keypad(pad, TRUE);
-    wgetch(pad);
-
-    mvwaddstr(pad, 35, 2, "This is displayed at line 35 in the pad");
-    mvwaddstr(pad, 40, 1, " Press any key to continue");
-    prefresh(pad,30,0,0,0,10,45);
-    keypad(pad, TRUE);
-    wgetch(pad);
-
-    delwin(pad);
-}
-
 static void Continue (WINDOW *win)
 {
+    noecho();
     wmove(win, 10, 1);
     mvwaddstr(win, 10, 1, " Press any key to continue");
     wrefresh(win);
@@ -696,11 +1341,16 @@ static void Continue (WINDOW *win)
 static void input_test(WINDOW *win)
 /* Input test, adapted from John Burnell's PDCurses tester */
 {
-    int w, h, bx, by, sw, sh, i, c,num;
-    char buffer [80];
+    int w, h, bx, by, sw, sh, i;
+
     WINDOW *subWin;
     wclear (win);
 
+#ifdef FOO
+    char buffer [80];
+    int num;
+#endif /* FOO */
+
     w  = win->_maxx;
     h  = win->_maxy;
     bx = win->_begx;
@@ -722,16 +1372,19 @@ static void input_test(WINDOW *win)
     wattrset(subWin, A_BOLD);
 #endif
     box(subWin, ACS_VLINE, ACS_HLINE);
+#ifdef FOO
+    mvwaddstr(subWin, 2, 1, "This is a subwindow");
+#endif /* FOO */
     wrefresh(win);
 
     nocbreak();
-    mvwaddstr(win, 2, 1, "Press some keys for 5 seconds");
-    mvwaddstr(win, 1, 1, "Pressing ^C should do nothing");
+    mvwaddstr(win, 1, 1, "Type random keys for 5 seconds.");
+    mvwaddstr(win, 2, 1,
+      "These should be discarded (not echoed) after the subwindow goes away.");
     wrefresh(win);
 
-    for (i = 0; i < 5; i++) {
-	werase (subWin);
-	box(subWin, ACS_VLINE, ACS_HLINE);
+    for (i = 0; i < 5; i++)
+    {
 	mvwprintw (subWin, 1, 1, "Time = %d", i);
 	wrefresh(subWin);
 	sleep(1);
@@ -750,27 +1403,18 @@ static void input_test(WINDOW *win)
     echo();
     wgetch(win);
     flushinp();
+    mvwaddstr(win, 12, 0,
+	      "If you see any key other than what you typed, flushinp() is broken.");
+    Continue(win);
 
     wmove(win, 9, 10);
     wdelch(win);
-    mvwaddstr(win, 4, 1, "The character should now have been deleted");
-    Continue(win);
-
-    wclear (win);
-    mvwaddstr(win, 2, 1, "Press a function key or an arrow key");
     wrefresh(win);
-    keypad(win, TRUE);
-    c = wgetch(win);
-
-    nodelay(win, TRUE);
-    wgetch(win);
-    nodelay(win, FALSE);
-
-    refresh();
-    wclear (win);
-    mvwaddstr(win, 3, 2, "The window should have moved");
-    mvwaddstr(win, 4, 2, "This text should have appeared without you pressing a key");
-    mvwprintw(win, 2, 2, "Keycode = %d", c);
+    wmove(win, 12, 0);
+    clrtoeol();
+    waddstr(win,
+	    "What you typed should now have been deleted; if not, wdelch() failed.");
+    Continue(win);
 
 #ifdef FOO
     /*
@@ -785,124 +1429,75 @@ static void input_test(WINDOW *win)
     Continue(win);
 }
 
-static void output_test (WINDOW *win)
+/****************************************************************************
+ *
+ * Menu test
+ *
+ ****************************************************************************/
+
+#define MENU_Y	4
+#define MENU_X	4
+
+static int input_virtualize(int c)
 {
-    WINDOW *win1;
-    char Buffer [80];
-    chtype ch;
-
-    wclear (win);
-    mvwaddstr(win, 1, 1, "You should now have a screen in the upper left corner, and this text should have wrapped");
-    mvwin(win, 2, 1);
-    Continue(win);
-
-	nl();
-    wclear(win);
-    mvwaddstr(win, 1, 1, "A new window will appear with this text in it");
-    mvwaddstr(win, 8, 1, "Press any key to continue");
-    wrefresh(win);
-    wgetch(win);
-
-    win1 = newwin(10, 50, 15, 25);
-    if(win1 == NULL)
-    {   endwin();
-        return;
-    }
-#ifdef A_COLOR
-    if (has_colors())
-    {
-	init_pair(3,COLOR_BLUE,COLOR_WHITE);
-	wattrset(win1, COLOR_PAIR(3));
-    }
+    if (c == '\n' || c == KEY_EXIT)
+	return(MAX_COMMAND + 1);
+    else if (c == 'n' || c == KEY_DOWN)
+	return(REQ_NEXT_ITEM);
+    else if (c == 'p' || c == KEY_UP)
+	return(REQ_PREV_ITEM);
     else
-	wattrset(win1, A_NORMAL);
-#else
-    wattrset(win1, A_NORMAL);
-#endif
-    wclear (win1);
-    mvwaddstr(win1, 5, 1, "This text should appear; using overlay option");
-    copywin(win, win1,0,0,0,0,10,50,TRUE);
+	return(c);
+}
 
-    box(win1,ACS_VLINE,ACS_HLINE);
-    wmove(win1, 8, 26);
-    wrefresh(win1);
-    wgetch(win1);
+static char *animals[] =
+{
+    "Lions", "Tigers", "Bears", "(Oh my!)", "Newts", "Platypi", "Lemurs",
+    (char *)NULL
+};
+static ITEM *items[sizeof(animals)/sizeof(char *)];
 
-    wclear(win1);
-    wattron(win1, A_BLINK);
-    mvwaddstr(win1, 4, 1, "This blinking text should appear in only the second window");
-    wattroff(win1, A_BLINK);
-    wrefresh(win1);
-    wgetch(win1);
-    delwin(win1);
+static void menu_test(void)
+{
+    MENU	*m;
+    ITEM	**ip = items;
+    char	**ap;
+    int		mrows, mcols;
+    WINDOW	*menuwin;
 
-    wclear(win);
-    wrefresh(win);
-    mvwaddstr(win, 6, 2, "This line shouldn't appear");
-    mvwaddstr(win, 4, 2, "Only half of the next line is visible");
-    mvwaddstr(win, 5, 2, "Only half of the next line is visible");
-    wmove(win, 6, 1);
-    wclrtobot (win);
-    wmove(win, 5, 20);
-    wclrtoeol (win);
-    mvwaddstr(win, 8, 2, "This line also shouldn't appear");
-    wmove(win, 8, 1);
-    wdeleteln(win);
-    Continue(win);
-    traceoff();
+    mvaddstr(LINES - 3, 0, "This is the menu test.");
 
-    wmove (win, 5, 9);
-    ch = winch (win);
+    for (ap = animals; *ap; ap++)
+	*ip++ = new_item(*ap, "");
+    *ip = (ITEM *)NULL;
 
-    wclear(win);
-    wmove (win, 6, 2);
-    waddstr (win, "The next char should be l:  ");
-    winsch (win, ch);
-    Continue(win);
+    m = new_menu(items);
 
-    wmove(win, 5, 1);
-    winsertln (win);
-    mvwaddstr(win, 5, 2, "The lines below should have moved down");
-    Continue(win);
+    scale_menu(m, &mrows, &mcols);
 
-#ifdef FOO
-    /*
-     * This test won't be portable until vsscanf() is
-     */
-    wclear(win);
-    wmove(win, 2, 2);
-    wprintw(win, "This is a formatted string in a window: %d %s\n", 42, "is it");
-    mvwaddstr(win, 10, 1, "Enter a string: ");
-    wrefresh(win);
-    echo();
-    wscanw (win, "%s", Buffer);
+    menuwin = newwin(mrows + 2, mcols +  2, MENU_Y, MENU_X);
+    set_menu_win(m, menuwin);
+    keypad(menuwin, TRUE);
+    box(menuwin, 0, 0);
 
-    wclear(win);
-    mvwaddstr(win, 10, 1, "Enter a string");
-    wrefresh(win);
-    clear();
-    move(0,0);
-    printw("This is a formatted string in stdscr: %d %s\n", 42, "is it");
-    mvaddstr(10, 1, "Enter a string: ");
-    refresh();
-    echo();
-    scanw ("%s", Buffer);
-#endif /* FOO */
+    set_menu_sub(m, derwin(menuwin, mrows, mcols, 1, 1));
 
-    wclear(win);
-    curs_set(2);
-    mvwaddstr(win, 1, 1, "The cursor should appear as a block");
-    Continue(win);
+    post_menu(m);
 
-    wclear(win);
-    curs_set(0);
-    mvwaddstr(win, 1, 1, "The cursor should have disappeared");
-    Continue(win);
+    while (menu_driver(m, input_virtualize(wgetch(menuwin))) != E_UNKNOWN_COMMAND)
+	continue;
 
-    wclear(win);
-    curs_set(1);
-    mvwaddstr(win, 1, 1, "The cursor should be an underline");
-    Continue(win);
+    (void) mvprintw(LINES - 2, 0,
+		     "You chose: %s\n", item_name(current_item(m)));
+    (void) addstr("Press any key to continue...");
+    wgetch(stdscr);
+
+    unpost_menu(m);
+    delwin(menuwin);
+
+    for (ip = items; *ip; ip++)
+	free_item(*ip);
+    free_menu(m);
 }
 
 /****************************************************************************
@@ -911,7 +1506,8 @@ static void output_test (WINDOW *win)
  *
  ****************************************************************************/
 
-bool do_single_test(const char c)
+static bool
+do_single_test(const char c)
 /* perform a single specified test */
 {
     switch (c)
@@ -926,19 +1522,16 @@ bool do_single_test(const char c)
 
     case 'c':
 	if (!has_colors())
-	    (void) printf("This %s terminal does not support color.\n",
-			  getenv("TERM"));
+	    Cannot("does not support color.");
 	else
 	    color_test();
 	return(TRUE);
 
     case 'd':
 	if (!has_colors())
-	    (void) printf("This %s terminal does not support color.\n",
-			  getenv("TERM"));
+	    Cannot("does not support color.");
 	else if (!can_change_color())
-	    (void) printf("This %s terminal has hardwired color values.\n",
-			  getenv("TERM"));
+	    Cannot("has hardwired color values.");
 	else
 	    color_edit();
 	return(TRUE);
@@ -951,10 +1544,18 @@ bool do_single_test(const char c)
 	acs_display();
 	return(TRUE);
 
+    case 'o':
+	demo_panels();
+	return(TRUE);
+
     case 'g':
 	acs_and_scroll();
 	return(TRUE);
 
+    case 'm':
+	menu_test();
+	return(TRUE);
+
     case 'p':
 	demo_pad();
 	return(TRUE);
@@ -963,10 +1564,6 @@ bool do_single_test(const char c)
 	input_test(stdscr);
 	return(TRUE);
 
-    case 'o':
-	output_test(stdscr);
-	return(TRUE);
-
     case '?':
 	(void) puts("This is the ncurses capability tester.");
 	(void) puts("You may select a test from the main menu by typing the");
@@ -983,7 +1580,7 @@ int main(const int argc, const char *argv[])
     char	buf[BUFSIZ];
 
     /* enable debugging */
-    traceon();
+    trace(TRACE_ORDINARY);
 
     /* tell it we're going to play with soft keys */
     slk_init(1);
@@ -1015,12 +1612,14 @@ int main(const int argc, const char *argv[])
 	(void) puts("e = exercise soft keys");
 	(void) puts("f = display ACS characters");
 	(void) puts("g = display windows and scrolling");
+	(void) puts("m = menu code test");
+	(void) puts("o = exercise panels library");
 	(void) puts("p = exercise pad features");
 	(void) puts("i = subwindow input test");
-	(void) puts("o = output test");
-	(void) puts("? = get help");
+	(void) puts("? = repeat this command summary");
 
 	(void) fputs("> ", stdout);
+	(void) fflush(stdout);		/* necessary under SVr4 curses */
 	(void) fgets(buf, BUFSIZ, stdin);
 
 	if (do_single_test(buf[0])) {
